// SPDX-FileCopyrightText: 2023 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

package org.pgpainless.key.generation

import openpgp.toArray
import openpgp.toPassphrase
import org.bouncycastle.openpgp.PGPSecretKeyRing
import org.pgpainless.algorithm.KeyFlag
import org.pgpainless.key.generation.type.KeyType
import org.pgpainless.key.generation.type.eddsa.EdDSACurve
import org.pgpainless.key.generation.type.rsa.RsaLength
import org.pgpainless.key.generation.type.xdh.XDHSpec
import org.pgpainless.key.protection.SecretKeyRingProtector
import org.pgpainless.util.Passphrase

class KeyRingTemplates {

    /**
     * Generate an RSA OpenPGP key consisting of an RSA primary key used for certification, a
     * dedicated RSA subkey used for signing and a third RSA subkey used for encryption.
     *
     * @param userId userId or null
     * @param length length of the RSA keys
     * @param passphrase passphrase to encrypt the key with. Can be empty for an unencrypted key.
     * @return key
     */
    @JvmOverloads
    fun rsaKeyRing(
        userId: CharSequence?,
        length: RsaLength,
        passphrase: Passphrase = Passphrase.emptyPassphrase()
    ): PGPSecretKeyRing =
        OpenPgpKeyGenerator()
            .buildV4Key()
            .fromTemplate()
            .composedRsa(
                userId = userId.toArray(),
                length = length,
                protector = SecretKeyRingProtector.unlockAnyKeyWith(passphrase))

    /**
     * Generate an RSA OpenPGP key consisting of an RSA primary key used for certification, a
     * dedicated RSA subkey used for signing and a third RSA subkey used for encryption.
     *
     * @param userId userId or null
     * @param length length of the RSA keys
     * @param password passphrase to encrypt the key with. Can be null or blank for unencrypted
     *   keys.
     * @return key
     */
    fun rsaKeyRing(
        userId: CharSequence?,
        length: RsaLength,
        password: CharSequence?
    ): PGPSecretKeyRing = rsaKeyRing(userId, length, password.toPassphrase())

    /**
     * Creates a simple RSA KeyPair of length {@code length} with user-id {@code userId}. The
     * KeyPair consists of a single RSA master key which is used for signing, encryption and
     * certification.
     *
     * @param userId user id.
     * @param length length in bits.
     * @param passphrase Password of the key. Can be empty for unencrypted keys.
     * @return {@link PGPSecretKeyRing} containing the KeyPair.
     */
    @JvmOverloads
    fun simpleRsaKeyRing(
        userId: CharSequence?,
        length: RsaLength,
        passphrase: Passphrase = Passphrase.emptyPassphrase()
    ): PGPSecretKeyRing =
        OpenPgpKeyGenerator()
            .buildV4Key()
            .fromTemplate()
            .singleRsa(
                userId = userId.toArray(),
                length = length,
                protector = SecretKeyRingProtector.unlockAnyKeyWith(passphrase))

    /**
     * Creates a simple RSA KeyPair of length {@code length} with user-id {@code userId}. The
     * KeyPair consists of a single RSA master key which is used for signing, encryption and
     * certification.
     *
     * @param userId user id.
     * @param length length in bits.
     * @param password Password of the key. Can be null or blank for unencrypted keys.
     * @return {@link PGPSecretKeyRing} containing the KeyPair.
     */
    fun simpleRsaKeyRing(userId: CharSequence?, length: RsaLength, password: CharSequence?) =
        simpleRsaKeyRing(userId, length, password.toPassphrase())

    /**
     * Creates a key ring consisting of an ed25519 EdDSA primary key and a X25519 XDH subkey. The
     * EdDSA primary key is used for signing messages and certifying the sub key. The XDH subkey is
     * used for encryption and decryption of messages.
     *
     * @param userId user-id
     * @param passphrase Password of the private key. Can be empty for an unencrypted key.
     * @return {@link PGPSecretKeyRing} containing the key pairs.
     */
    @JvmOverloads
    fun simpleEcKeyRing(
        userId: CharSequence?,
        passphrase: Passphrase = Passphrase.emptyPassphrase()
    ): PGPSecretKeyRing =
        OpenPgpKeyGenerator()
            .buildV4Key()
            .setPrimaryKey(
                KeyType.EDDSA(EdDSACurve._Ed25519),
                listOf(KeyFlag.CERTIFY_OTHER, KeyFlag.SIGN_DATA)) {
                    if (userId != null) {
                        addUserId(userId)
                    }
                }
            .addEncryptionSubkey(KeyType.XDH(XDHSpec._X25519))
            .build(passphrase)

    /**
     * Creates a key ring consisting of an ed25519 EdDSA primary key and a X25519 XDH subkey. The
     * EdDSA primary key is used for signing messages and certifying the sub key. The XDH subkey is
     * used for encryption and decryption of messages.
     *
     * @param userId user-id
     * @param password Password of the private key. Can be null or blank for an unencrypted key.
     * @return {@link PGPSecretKeyRing} containing the key pairs.
     */
    fun simpleEcKeyRing(userId: CharSequence?, password: CharSequence?): PGPSecretKeyRing =
        simpleEcKeyRing(userId, password.toPassphrase())

    /**
     * Generate a modern PGP key ring consisting of an ed25519 EdDSA primary key which is used to
     * certify an X25519 XDH encryption subkey and an ed25519 EdDSA signing key.
     *
     * @param userId primary user id
     * @param passphrase passphrase for the private key. Can be empty for an unencrypted key.
     * @return key ring
     */
    @JvmOverloads
    fun modernKeyRing(
        userId: CharSequence?,
        passphrase: Passphrase = Passphrase.emptyPassphrase()
    ): PGPSecretKeyRing =
        OpenPgpKeyGenerator()
            .buildV4Key()
            .fromTemplate()
            .ed25519Curve25519(
                userId = userId.toArray(),
                protector = SecretKeyRingProtector.unlockAnyKeyWith(passphrase))

    /**
     * Generate a modern PGP key ring consisting of an ed25519 EdDSA primary key which is used to
     * certify an X25519 XDH encryption subkey and an ed25519 EdDSA signing key.
     *
     * @param userId primary user id
     * @param password passphrase for the private key. Can be null or blank for an unencrypted key.
     * @return key ring
     */
    fun modernKeyRing(userId: CharSequence?, password: CharSequence?): PGPSecretKeyRing =
        modernKeyRing(userId, password.toPassphrase())
}
